import sys
import signal
from enum import Enum
from os import getppid, kill
from queue import Empty
from copy import deepcopy
from multiprocessing import Process

from cml.shared.logger import ConstructionInfo


class BackgroundJobStatus(Enum):
    DONE = 1


class BackgroundConstructor(Process):
    def __init__(self, models, settings, input_queue, output_queue, control_queue):
        super(BackgroundConstructor, self).__init__(name="BackgroundConstructor")
        self.models = models
        self.settings = settings
        self.input_queue = input_queue
        self.output_queue = output_queue
        self.control_queue = control_queue

    def run(self):

        while 1:
            try:
                learnblock = self.input_queue.get(timeout=0)

                for args in self.prepare_args(learnblock):
                    logger_info = ConstructionInfo()
                    learnblock, model, _ = args
                    if learnblock.labeled():
                        logger_info.state = ConstructionInfo.State.PASSED
                        self.output_queue.put((logger_info, learnblock))

                    else:
                        trained_model = model.train(learnblock)

                        if model.cluster_specific:
                            if model.found_cluster != model.cluster: continue
                        else:
                            if model.found_cluster <= 1: continue

                        for cluster, size in trained_model.cluster_sizes.items():
                            if size < self.settings.min_category_size:
                                break
                        else:
                            labeled = self.label_block(learnblock, trained_model)

                            # Fill the logger object
                            logger_info.n_cluster = trained_model.found_cluster
                            logger_info.subject = trained_model.abbreviation
                            logger_info.state = ConstructionInfo.State.CONSTRUCTED
                            self.output_queue.put((logger_info, labeled))

                if logger_info.state is None:
                    logger_info.state = ConstructionInfo.State.DISCARDED
                    self.output_queue.put((logger_info, None))

                self.output_queue.put(BackgroundJobStatus.DONE)

            except Empty:
                pass
                # sys.exit()

            except Exception as e:
                parent = getppid()
                kill(parent, signal.SIGALRM)
                sys.exit(-1)

    def label_block(self, block, trained_model):
        copied_block = deepcopy(block)
        labels = trained_model.get_labels()
        copied_block.set_column("Z", labels)
        copied_block.n_cluster = trained_model.found_cluster
        copied_block.origin = block.origin
        subject = "{}{:02}".format(trained_model.abbreviation, trained_model.found_cluster)
        copied_block.subject = subject
        return copied_block

    def prepare_args(self, learnblock):
        for ml_model in self.models:
            if ml_model.cluster_specific:
                for complexity in range(2, self.settings.max_categories+1):
                    ml_model.cluster = complexity
                    # yield deepcopy(learnblock), deepcopy(ml_model), complexity
                    yield deepcopy(learnblock), ml_model, complexity
            else:
                yield deepcopy(learnblock), ml_model, None
                # yield (deepcopy(learnblock), deepcopy(ml_model), None)
