import pickle
from bisect import bisect
from collections import deque
from collections import defaultdict

from cml.domain.reconstruction import PragmaticMachineLearningModel


TS_QUEUE = deque()


def notify_inverted_index(func):
    def wrapper(self, *args, **kwargs):
        for obs in self.observer:
            getattr(obs, func.__name__)(*args, **kwargs)
        return func(self, *args, **kwargs)
    return wrapper


class KnowledgeDatabase:
    def __init__(self, highest_tier, build_block_from_rows):
        self.database = [KnowledgeDomain(i) for i in range(highest_tier)]
        self.observer = []
        self.source = KnowledgeSource(build_block_from_rows)
        self.n_models = 0

    def __str__(self):
        return str(self.database)

    def __contains__(self, item):
        if not isinstance(item, PragmaticMachineLearningModel):
            raise TypeError()
        try:
            self.get(item.uid)
            return True
        except KeyError:
            return False

    def extend(self, block):
        return self.source.append(block)

    def get_block(self, indices):
        return self.source.get(indices)

    @notify_inverted_index
    def insert(self, model):
        if model not in self:
            self.database[model.tier].insert(model)
            self.n_models += 1

    @notify_inverted_index
    def remove(self, model):
        self.database[model.tier].remove(model)
        self.n_models -= 1

    @notify_inverted_index
    def replace(self, replaced, replacer):
        self.database[replaced.tier].replace(replaced, replacer)

    def get(self, uid: str):
        _, tier, _ = uid.split(".")
        return self.database[int(tier)].get(uid)

    def dump(self, path: str):
        file = open(path, "wb")
        pickle.dump(self, file)
        file.close()

    def model_counter(self):
        def counts(tier):
            return self.database[tier].counter()
        return counts

    def remove_dependent_models(self, relative_model):
        to_remove = []
        for domain in self.database:
            for model in domain.knowledge.values():
                # TODO (dmt): Fix origin datatype.
                if isinstance(model.origin, str):
                    if model.origin == relative_model.uid:
                        to_remove.append(model)
                elif isinstance(model.origin, list):
                    if relative_model.uid in model.origin:
                        to_remove.append(model)
                elif isinstance(model.origin, tuple):
                    if relative_model.uid in model.origin:
                        to_remove.append(model)

                else: continue

        return to_remove
        # for model in to_remove:
        #    self.remove(model)

    # def last_index_source(self):
    #     if self.sources:
    #         last_data_source = self.sources[-1]
    #         return last_data_source.source.indices()[-1]
    #     else:
    #         return 0
    #
    # def insert_source(self, source):
    #     self.sources.append(source)
    #


class KnowledgeDomain:
    def __init__(self, tier):
        self.tier = tier
        self.knowledge = {}
        self.biggest_id = 0

    def __str__(self):
        return "\n".join(["{:~^100}".format(self.tier), "{}".format(
            self.knowledge)])

    def get(self, uid):
        return self.knowledge[uid]

    def insert(self, model):
        self.knowledge[model] = model
        #self.update_biggest_id(model.identifier)

    def remove(self, model):
        del self.knowledge[model]

    def replace(self, replaced, replacer):
        del self.knowledge[replaced]
        self.knowledge[replacer] = replacer
        #self.update_biggest_id(replacer.identifier)

    def counter(self):
        self.biggest_id += 1
        return self.biggest_id


class KnowledgeSource:
    def __init__(self, pandas_factory):
        self.source = {}
        self.n_block = 0
        self.pandas_factory = pandas_factory
        self.__last_index = 0
        self.__search_ranges = []

    def append(self, block):
        self.source[self.n_block] = block
        self.__reindex(block)
        self.__update_search_range()
        self.n_block += 1
        return self.n_block - 1

    def get(self, indices):
        rows = list(self.indices_to_rows(indices))
        return self.pandas_factory(rows)

    def __reindex(self, block):
        new_end_indx = self.__last_index+block.n_rows()
        block.reindex(range(self.__last_index, new_end_indx))
        self.__last_index = new_end_indx

    def __update_search_range(self):
        self.__search_ranges.append(self.__last_index)

    def indices_to_rows(self, indices):
        for index in indices:
            try:
                yield self.source[
                    self.bucket_search(index)].row_as_df_via_index(index)
            except (IndexError, KeyError) as error:
                print("Index: ", index)
                print(self.__search_ranges)
                bucket = self.bucket_search(index)
                print(bucket)
                # print(self.source[bucket].indices())
                print(self.source[0].indices())
                print(self.source[1].indices())
                print(self.source[2].indices())
                print(self.source[3].indices())
                print(self.n_block)

                raise error

    def bucket_search(self, index):
        return bisect(self.__search_ranges, index,
                      lo=0, hi=self.n_block-1)


class RelativeFinder:
    def __init__(self, tiers):
        self.indices = {}
        for tier in range(tiers):
            self.indices[tier] = (
                defaultdict(set),
                defaultdict(set),
                defaultdict(set))

    def find(self, pair, model, deconst_full_tolerance):
        index_t, index_z, index_sigma = self.get_index_tier(model)
        if pair == ("T", "Z"):
            set_t = index_t[(model.min_timestamp, model.max_timestamp)]
            set_z = index_z[model.purpose]
            relatives = set_t.intersection(set_z)
        elif pair == ("T", "Sigma"):
            set_t = index_t[(model.min_timestamp, model.max_timestamp)]
            set_s = index_sigma[model.subject]
            relatives = set_t.intersection(set_s)
        elif pair == ("Sigma", "Z"):
            set_s = index_sigma[model.subject]
            set_z = index_z[model.purpose]
            relatives = set_s.intersection(set_z)
        elif pair == ("complete", ):
            set_t = set()
            for min_time, max_time in index_t.keys():
                if self.complete_relationship_constraint(model, min_time,
                    max_time, deconst_full_tolerance):
                    set_t.update(index_t[(min_time, max_time)])
            set_s = index_sigma[model.subject]
            set_z = index_z[model.purpose]
            relatives = set_t.intersection(set_s).intersection(set_z)
        else: raise ValueError()

        if model in relatives:
            relatives.remove(model)

        return relatives

    def complete_relationship_constraint(self, model, min_t, max_t, deconst_full_tolerance):
        return self.__time_exclusion(model, min_t, max_t, deconst_full_tolerance) or \
               self.__time_inclusion(model, min_t, max_t, deconst_full_tolerance)

    def __time_inclusion(self, model, min_t, max_t, tolerance):
        return model.min_timestamp >= min_t and model.max_timestamp <= max_t and \
               model.min_timestamp - min_t <= model.min_timestamp * tolerance and \
               model.max_timestamp - max_t * (-1) <= model.max_timestamp * tolerance

    def __time_exclusion(self, model, min_t, max_t, tolerance):
        return model.min_timestamp <= min_t and model.max_timestamp >= max_t and \
               model.min_timestamp - min_t * (-1) <= model.min_timestamp * tolerance and \
               model.max_timestamp - max_t <= model.max_timestamp * tolerance

    def remove(self, model):
        t_list, s_list, z_list = self.get_index_lists(model,
                                                      time=True,
                                                      subject=True,
                                                      purpose=True)
        s_list.remove(model)
        z_list.remove(model)
        t_list.remove(model)

    def replace(self, replaced, replacer):
        t_list, s_list, z_list = self.get_index_lists(replaced,
                                                      time=True,
                                                      subject=True,
                                                      purpose=True)
        t_list.remove(replaced)
        s_list.remove(replaced)
        z_list.remove(replaced)
        t, z, sigma = self.get_index_tier(replacer)
        t[(replacer.min_timestamp, replacer.max_timestamp)].add(replacer)
        z[replacer.purpose].add(replacer)
        sigma[replacer.subject].add(replacer)

    def insert(self, model):
        t, z, sigma = self.get_index_tier(model)
        t[(model.min_timestamp, model.max_timestamp)].add(model)
        z[model.purpose].add(model)
        sigma[model.subject].add(model)

    def get_index_lists(self, model, time, subject, purpose):
        t, z, sigma = self.get_index_tier(model)
        t_list = s_list = z_list = None
        if time:
            t_list = t[(model.min_timestamp, model.max_timestamp)]
        if subject:
            s_list = sigma[model.subject]
        if purpose:
            z_list = z[model.purpose]
        return t_list, s_list, z_list

    def get_index_tier(self, model):
        index_t = self.indices[model.meta.knowledge_tier][0]
        index_z = self.indices[model.meta.knowledge_tier][1]
        index_sigma = self.indices[model.meta.knowledge_tier][2]
        return index_t, index_z, index_sigma

    def extend(self): pass
