import re
from random import seed, shuffle
from collections import defaultdict, namedtuple
from functools import partial
from multiprocessing import Pool, cpu_count

from krippendorff import alpha

from cml.shared.errors import (
    ReconstructionFailure,
    NotEnoughFeaturesWarning)

InterimPragmatic = namedtuple("InterimPragmatic",
                              "model, predictions, learnblock")


class Metadata:
    def __init__(self, knowledge_domain, knowledge_tier, identifier, pre_image,
                 pre_image_features, pre_image_labels, t_min, t_max, sigma,
                 zeta):
        """
        Sigma = ANN, RF
        Zeta: C.2.Kme02
        """
        assert isinstance(sigma, tuple)

        self.knowledge_domain = knowledge_domain
        self.knowledge_tier = knowledge_tier
        self.identifier = identifier
        self.pre_image = pre_image
        self.pre_image_features = pre_image_features
        self.pre_image_labels = pre_image_labels
        self.t_min = t_min
        self.t_max = t_max
        self.sigma = sigma
        self.zeta = zeta

    def __str__(self):
        return f"Knowledge domain: <{self.knowledge_domain}> \n" \
               f"Knowledge tier: <{self.knowledge_tier}> \n" \
               f"Identifier: <{self.identifier}> \n" \
               f"Pre image: <{self.pre_image}> \n" \
               f"Pre image labels: <{self.pre_image_labels}> \n" \
               f"Pre image features: <{self.pre_image_features}> \n" \
               f"T min: <{self.t_min}> \n" \
               f"T max: <{self.t_max}> \n" \
               f"Subjects: <{self.sigma}> \n" \
               f"Purposes: <{self.zeta}> \n"

    def __hash__(self):
        return hash(".".join([self.knowledge_domain,
                              str(self.knowledge_tier),
                              str(self.identifier)]))

    def __eq__(self, other):
        if not isinstance(other, Metadata):
            raise TypeError()

        return hash(self) == hash(other)


class PragmaticMachineLearningModel:

    def __init__(self, meta, model, learnblock):
        self.meta = meta
        self.model = model
        self.origin = learnblock.origin
        self.__block_to_features = {}
        self.__block_to_indices = {}

        regex = "\("+str(self.meta.knowledge_domain)+"\.\d+\.\d+"
        if any([re.match(regex, str(o)) for o in learnblock.origin]):
            self.learnblock = learnblock
        else:
            self.learnblock = None

        assert isinstance(self.origin, tuple)

    def __str__(self) -> str:
        return self.uid

    def __repr__(self) -> str:
        return self.uid

    def __hash__(self) -> int:
        return hash(self.meta)

    def __eq__(self, other):
        if isinstance(other, PragmaticMachineLearningModel):
            return hash(self.meta) == hash(other.meta)

        if isinstance(other, str):
            return hash(self) == hash(other)

        raise NotImplementedError()

    @property
    def domain(self): return self.meta.knowledge_domain

    @property
    def tier(self):
        return self.meta.knowledge_tier

    @property
    def identifier(self): return self.meta.identifier

    @identifier.setter
    def identifier(self, value): self.meta.identifier = value

    @property
    def min_timestamp(self): return self.meta.t_min

    @property
    def max_timestamp(self): return self.meta.t_max

    @property
    def pre_image_features(self): return self.meta.pre_image_features

    @property
    def pre_image(self): return self.meta.pre_image

    @property
    def pre_image_labels(self): return self.meta.pre_image_labels

    @property
    def subject(self): return self.meta.sigma

    @property
    def purpose(self): return self.meta.zeta

    @property
    def uid(self):
        uid = "." .join([self.domain, str(self.tier), str(self.identifier)])
        return uid

    @property
    def accuracy(self): return self.model.accuracy

    @property
    def reliability(self): return self.model.reliability

    @reliability.setter
    def reliability(self, value):
        self.model.reliability = value

    def fusion(self, model, knowledge_tier, new_identifier):
        return Metadata(knowledge_domain=self.meta.knowledge_domain,
                        knowledge_tier=knowledge_tier,
                        identifier=new_identifier,
                        pre_image=None,
                        pre_image_features=None,
                        pre_image_labels=None,
                        t_min=min(self.meta.t_min, model.min_timestamp),
                        t_max=max(self.meta.t_max, model.max_timestamp),
                        sigma=tuple(set(self.subject + model.subject)),
                        zeta=tuple(set(self.meta.zeta + model.purpose)))

    def trained_with(self, knowledge):
        if self.learnblock:
            return self.learnblock

        learnblock = knowledge.get_block(self.pre_image)
        preserved_features = self.pre_image_features + ["T", "Zigma", "Z"]
        for feature in learnblock.columns():
            if feature not in preserved_features:
                learnblock.drop_columns([feature])
        learnblock.set_column("Z", self.pre_image_labels)
        learnblock.subject = self.subject
        learnblock.origin = self.origin
        return learnblock

        print("TRAINED WITH")
        print(block.as_numpy_array())
        print(block.df.loc[self.pre_image])
        learnblock = block.new_block_from(self.pre_image,
                                          index=True)
        print(learnblock.as_numpy_array())
        preserved_features = self.pre_image_features + ["T", "Zigma", "Z"]
        for feature in learnblock.columns():
            if feature not in preserved_features:
                learnblock.drop_columns([feature])
        print(learnblock.as_numpy_array())
        learnblock.set_column("Z", self.pre_image_labels)
        learnblock.subject = self.subject
        print(learnblock.as_numpy_array())
        print("TRAINED WITH")
        return learnblock


class SubprocessDTO:
    def __init__(self, count, model, learnblock, tier, abv, settings):
        self.count = count
        self.model = model
        self.learnblock = learnblock
        self.tier = tier
        self.abv = abv
        self.settings = settings


class Reconstructor:
    CONCEPTUAL_KNOWLEDGE_ABBREVIATION = "C"
    PROCEDURAL_KNOWLEDGE_ABBREVIATION = "P"

    def __init__(self, settings, ml_models, mode):
        self.mode = mode
        self.settings = settings
        self.ml_models = ml_models
        self.model_counter = None
        self.__reconstruction = self._init_reconstruction(mode)

    def _init_reconstruction(self, mode):
        if mode == "conceptual":
            return partial(self.reconstruct_conceptual, krippen="nominal")

    def inject_counter(self, counter):
        self.model_counter = counter

    def reconstruct_parallel(self, tier, learnblock, models=None, meta=None):
        from domain.reconstruction.reconstruction_parallel import reconstruct_in_parallel
        from domain.reconstruction.reconstruction_parallel import set_valid_counter

        if not models:
            args = []
            for count, model in enumerate(self.ml_models, 1):
                args.append(
                    SubprocessDTO(
                        count, model, learnblock, tier,
                        self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION,
                        self.settings))

            set_valid_counter(len(self.ml_models))
            pool = Pool(cpu_count()-2)
            interims = pool.map(reconstruct_in_parallel, args)
            pool.close()
            pool.join()
            reliability = calc_intersubject_reliability(interims)
            for interim in interims:
                interim.model.identifier = self.model_counter(tier)
                interim.model.reliability = reliability
            if not interims: raise ReconstructionFailure

            return reliability, [(interim.model, interim.learnblock)
                                 for interim in interims]



    def reconstruct_serial(self, tier, lb, which_models=None, meta=None):
        if not (lb.n_columns(effective=True) >= 2):
            raise NotEnoughFeaturesWarning()

        which_models = list(self.get_models(which_models))
        interims = list(self.__reconstruction(tier, lb, which_models, meta=meta))
        reliability = calc_intersubject_reliability(interims)
        for interim in interims:
            interim.model.reliability = reliability

        return reliability, [(interim.model, interim.learnblock)
                             for interim in interims]

    def reconstruct_conceptual(self, tier, lb, which_models, krippen=None,
                               meta=None):
        for model in which_models:
            train, eval_ = split(lb, self.settings.reliability_sample)
            trained = self.fit_model(train, model)
            if valid_reconstructed(trained, "conceptual", self.settings):
                predictions = trained.predict(eval_.as_numpy_array())
                metadata = self.get_metadata(tier, lb, trained, meta)
                pragmatic = PragmaticMachineLearningModel(metadata, trained, lb)
                yield InterimPragmatic(pragmatic, predictions, lb)


                #reliability = calc_reliability(trained, eval_, krippen)
                #if reliability >= self.settings.min_reliability:
                #    trained.reliability = reliability
                #    metadata = self.get_metadata(tier, lb, trained, meta)
                #    pragmatic = PragmaticMachineLearningModel(
                #        metadata, trained, lb)
                #    reliabilities[reliability].append(Pair(pragmatic, lb))
        # return reliabilities

    def get_models(self, which_models):
        if not which_models:
            for model in self.ml_models:
                yield model
        else:
            for model in self.ml_models:
                if model.subject in which_models:
                    yield model

    def fit_model(self, train, model):
        data = train.as_numpy_array()
        labels = [label for label in train.get_column("Z")]
        return model.train(data, labels)

    def get_metadata(self, tier, lb, model, meta=None):
        if not meta:
            zeta = ".".join([
                self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION,
                str(tier),
                lb.subject
            ])
            metadata = Metadata(self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION,
                                tier,
                                self.model_counter(tier),
                                lb.indices(),
                                lb.columns(),
                                lb.get_column("Z"),
                                lb.min_timestamp(),
                                lb.max_timestamp(),
                                (model.subject,),
                                (zeta,))
        else:
            metadata = meta
            metadata.pre_image_features = lb.columns()
            metadata.pre_image_labels = lb.get_column("Z")
            metadata.pre_image = lb.indices()

        return metadata


def split(learnblock, reliability_sample):
    seed(7)
    indices = learnblock.indices()
    eval_size = int(learnblock.n_rows() * reliability_sample)
    # eval_idx = sample(indices, eval_size)
    # shuffle(indices)
    eval_idx, train_idx = indices[:eval_size], indices[eval_size+1:]

    # train_idx = list(set(indices).difference(set(eval_idx)))
    return learnblock.new_block_from(train_idx, index=True), \
        learnblock.new_block_from(eval_idx, index=True)


def calc_intersubject_reliability(interims):
    predictions = [interim.predictions for interim in interims]
    if len(predictions) == 1:
        return 1.0

    reliability_alpha = alpha(predictions)
    return float(reliability_alpha)


def calc_reliability(trained_model, eval_block, metric):
    y_pre = trained_model.predict(eval_block.as_numpy_array())
    y_true = [i for i in eval_block.get_column("Z")]
    reliability_data = [y_pre, y_true]
    return alpha(reliability_data, level_of_measurement=metric)


def valid_reconstructed(model, knowledge_domain, settings):
    if knowledge_domain == "conceptual":
        return model.accuracy >= settings.min_test_accuracy
    # else:
    #    return model.mean_error <= max_test_error_avg and \
    #        model.max_error <= max_test_error_max

