from os.path import join
import multiprocessing
from collections import defaultdict
from copy import deepcopy
from logging import getLogger, FileHandler, StreamHandler
from contextlib import suppress

from cml.shared.parameter import PROTOCOL_LEVEL
from cml.usecases.usecase import Usecase
from cml.domain.knowledge import TS_QUEUE
from domain.knowledge import KnowledgeDatabase
from domain.reconstruction.selection import determine_winner
from domain.preprocessing import Source, LearnblockIdentifier, Preprocessor
from cml.shared.errors import (
    LearnblockIdentifikationFailure,
    ReconstructionFailure,
    WinnerSelectionFailure,
    SelectionFailure,
    ModeError)
from cml.shared.logger import (
    StartLogging, BlockLogging, FinishLogging,
    FirstLevelLogger, HighLevelLogger,
    WinnerLogger,
    LearnblockLogger,
    ReducerLogger,
    ConstLogger,
    ReconstLogger,
    DeconstLogger)


class KnowledgeSearcher:
    def __init__(self, source_wrapper, constructor_wrapper,
                 feature_selector_wrapper, reconstructor_wrapper,
                 winner_selector_wrapper, deconstructor_wrapper,
                 knowledge_database, stdout, debug, parallel):

        self.source_wrapper = source_wrapper
        self.constructor_wrapper = constructor_wrapper
        self.feature_selector_wrapper = feature_selector_wrapper
        self.reconstructor_wrapper = reconstructor_wrapper
        self.winner_selector_wrapper = winner_selector_wrapper
        self.deconstructor_wrapper = deconstructor_wrapper
        self.knowledge_database = knowledge_database
        self.__stdout = stdout
        self.__debug = debug
        self.__parallel = parallel
        self.log_destination = None

    @StartLogging
    def __enter__(self):
        logger = self.__get_logger()
        self.__prepare_logger(logger)
        self.__prepare_counter()
        # Todo copy the knowledge
        return self

    def search(self, block):
        with suppress(LearnblockIdentifikationFailure):
            for level, lb in self.source_wrapper.learnblocks(block):
                reconstructed = defaultdict(list)

                try:
                    for _, labeled in self.constructor_wrapper.construct(lb):
                        reduced = self.feature_selector_wrapper.reduce(labeled)
                        reliability, pragmatics = \
                            self.reconstructor_wrapper.reconstruct(level, reduced)
                        reconstructed[reliability].extend(pragmatics)

                except (SelectionFailure, ReconstructionFailure):
                    continue

                try:
                    winner, winner_lb = self.winner_selector_wrapper.determine(
                        reconstructed)
                    self.deconstructor_wrapper.deconstruct(
                        level, winner, winner_lb, self.knowledge_database)
                except WinnerSelectionFailure:
                    continue

        return deepcopy(self.knowledge_database), self.source_wrapper.halde

    @FinishLogging
    def __exit__(self, exc_type, exc_val, exc_tb):
        # Save knowledge database
        self.__close_all_children()
        pass

    def __get_logger(self):
        iteration_log_file = join(self.log_destination, "iteration.log")
        iteration_handler = FileHandler(iteration_log_file)
        iteration_handler.level = PROTOCOL_LEVEL
        iteration_logger = getLogger("iterationLogger")

        # Remove loggers from other sessions
        if iteration_logger.handlers: iteration_logger.handlers.clear()
        iteration_logger.addHandler(iteration_handler)

        if self.__stdout:
            iteration_stdout_handler = StreamHandler()
            iteration_stdout_handler.level = PROTOCOL_LEVEL
            iteration_logger.addHandler(iteration_stdout_handler)

        if self.__debug:
            pass

        return iteration_logger

    def __prepare_logger(self, logger):
        self.source_wrapper.logger = logger
        self.constructor_wrapper.logger = logger
        self.feature_selector_wrapper.logger = logger
        self.reconstructor_wrapper.logger = logger
        self.winner_selector_wrapper.logger = logger
        self.deconstructor_wrapper.logger = logger

    def __prepare_counter(self):
        model_counter = self.source_wrapper.model_counter()
        self.reconstructor_wrapper.prepare_counter(model_counter)

    def __close_all_children(self):
        if self.__parallel:
            for child in multiprocessing.active_children():
                child.kill()


class KnowledgeSearcherUsecase(Usecase):
    def process(self, request):
        if not self.valid_modes(request):
            raise ModeError()

        parts = self.wrapp_components(request)
        return KnowledgeSearcher(*parts,
                                 stdout=request.stdout,
                                 parallel=request.parallel,
                                 debug=request.debug)

    def wrapp_components(self, request):
        block_wrapper, knowledge, identifier = self.build_components(request)
        identifier_wrapper = LearnblockIdentifierWrapper(identifier)
        source_w = SourceWrapper(block_wrapper, identifier_wrapper)
        const_w = ConstructorWrapper(request.constructor)
        select_w = FeatureSelectorWrapper(request.feature_selector)
        recon_w = ReconstructorWrapper(request.reconstructor)
        winner_determ_w = WinnerDeterminatorWrapper(determine_winner)

        self.inject_reconstructor(request)
        decon_w = DeconstructorWrapper(request.deconstructor)

        self.set_processing_type(const_w, recon_w, decon_w, request.parallel)
        return (source_w,
                const_w,
                select_w,
                recon_w,
                winner_determ_w,
                decon_w,
                knowledge)

    def build_components(self, request):
        knowledge = self.create_database(request)
        identifier = LearnblockIdentifier(request.block_processing_settings,
                                          request.density_estimator)
        preprocessor = Preprocessor(request.preprocessor_settings)
        # TODO (dmt): User should provide knowledge database
        source = Source(knowledge, preprocessor, request.df_converter)
        block_wrapper = BlockProcessorWrapper(source)
        return block_wrapper, knowledge, identifier

    def inject_reconstructor(self, request):
        request.deconstructor.reconstructor = request.reconstructor

    def set_processing_type(self, const_w, recon_w, decon_w, parallel):
        const_w.parallel_processing = parallel
        recon_w.parallel_processing = parallel
        decon_w.parallel_processing = parallel

    def valid_modes(self, request):
        return request.constructor.mode == request.reconstructor.mode == \
               request.deconstructor.mode

    def create_database(self, request):
        knowledge = KnowledgeDatabase(
            request.general_settings.highest_tier,
            request.build_block_from_rows) \
            if request.knowledge is None else request.knowledge
        knowledge.observer.append(request.deconstructor.relative_finder)
        return knowledge


class Wrapper:
    def __init__(self):
        self.logger = None


class BlockProcessorWrapper(Wrapper):
    def __init__(self, source):
        super().__init__()
        self.source = source

    @BlockLogging
    def process(self, block):
        return self.source.process(block)

    def get_knowledge(self):
        return self.source.knowledge


class SourceWrapper:
    def __init__(self, block_wrapper, identifier_wrapper):
        self.block_wrapper = block_wrapper
        self.identifier_wrapper = identifier_wrapper
        self._logger = None

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger):
        self._logger = logger
        self.identifier_wrapper.logger = logger

    @property
    def knowledge(self):
        return self.block_wrapper.get_knowledge()

    @property
    def halde(self):
        return self.identifier_wrapper.halde

    def learnblocks(self, block):
        for level, lb, in self.tier_one_learnblock(block):
            yield level, lb

        for level, lb in self.tier_n_learnblocks():
            yield level, lb

    @FirstLevelLogger
    def tier_one_learnblock(self, block):
        level = 1
        pandas_block = self.block_wrapper.process(block)
        lb = self.identifier_wrapper.identify(pandas_block)
        yield level, lb

    @HighLevelLogger
    def tier_n_learnblocks(self):
        while TS_QUEUE:
            tier, lb = TS_QUEUE.pop()
            if tier and lb:
                yield tier, lb

    def model_counter(self):
        return self.knowledge.model_counter()


class LearnblockIdentifierWrapper(Wrapper):
    def __init__(self, identifier):
        super().__init__()
        self.identifier = identifier
        self._halde = None

    @property
    def halde(self):
        return self._halde

    @LearnblockLogger
    def identify(self, block):
        learnblock, halde = self.identifier.identify(block)
        self._halde = halde
        return learnblock


class ConstructorWrapper(Wrapper):
    def __init__(self, constructor):
        super().__init__()
        self.constructor = constructor
        self.act = self.constructor.construct_serial
        self._parallel_processing = False

    @property
    def parallel_processing(self):
        return self._parallel_processing

    @parallel_processing.setter
    def parallel_processing(self, parallel):
        if parallel:
            self.act = self.constructor.construct_parallel
        else:
            self.act = self.constructor.construct_serial

    @ConstLogger
    def construct(self, lb):
        for log_info, labeled in self.act(lb):
            yield log_info, labeled


class FeatureSelectorWrapper(Wrapper):
    def __init__(self, selector):
        super().__init__()
        self.selector = selector

    def prepare(self):
        pass

    @ReducerLogger
    def reduce(self, labeled):
        return self.selector.select(labeled)


class ReconstructorWrapper(Wrapper):
    def __init__(self, reconstructor):
        super().__init__()
        self.reconstructor = reconstructor
        self.act = reconstructor.reconstruct_serial
        self._parallel_processing = False

    @property
    def parallel_processing(self):
        return self._parallel_processing

    @parallel_processing.setter
    def parallel_processing(self, paralll):
        if paralll:
            self.act = self.reconstructor.reconstruct_parallel
        else:
            self.act = self.reconstructor.reconstruct_serial

    @ReconstLogger
    def reconstruct(self, tier, reduced):
        return self.act(tier, reduced)

    def prepare_counter(self, model_counter):
        self.reconstructor.model_counter = model_counter


class WinnerDeterminatorWrapper(Wrapper):
    def __init__(self, func):
        super().__init__()
        self.func = func

    @WinnerLogger
    def determine(self, reliabilities):
        return self.func(reliabilities)


class DeconstructorWrapper(Wrapper):
    def __init__(self, deconstructor):
        super().__init__()
        self.deconstructor = deconstructor
        self.act = deconstructor.deconstruct_serial
        self._parallel_processing = False

    @property
    def parallel_processing(self):
        return self._parallel_processing

    @parallel_processing.setter
    def parallel_processing(self, paralll):
        if paralll:
            self.act = self.deconstructor.deconstruct_serial
        else:
            self.act = self.deconstructor.deconstruct_parallel

    @DeconstLogger
    def deconstruct(self, tier, pragmatic, reduced, knowledge):
        for log_info in self.act(tier, pragmatic, reduced, knowledge):
            yield log_info


class BlockGeneratorWrapper(Wrapper):
    def __init__(self, source):
        super().__init__()
        self.source = source

    def prepare(self, old_source_size, start_index, end_index):
        self.source.n_source = old_source_size
        self.source.reindex(start_index, end_index)
        self.source.remap_feature_names()

    def blocks(self):
        for block in self.source.blocks:
            yield block

    def update_generator(self, lb):
        self.source._flip_source_halde_flags(lb.indices())

    def get_origin(self):
        return self.source.n_source

