from unittest.mock import MagicMock, patch
from collections import Counter

import pytest
from sklearn.cluster.spectral import SpectralClustering
from sklearn.cluster.k_means_ import KMeans

import cml
from cml.ports.source_adapter import PandasBlock
from cml.shared.settings import specific_settings_factory
from cml.domain.construction import Constructor
from cml.ports.ml_adapter import MachineLearningModel


@pytest.fixture(scope='module')
def conceptual_constructor():
    algorithms = [
        ("Kme", KMeans(algorithm="elkan")),
        ("Spe", SpectralClustering(affinity="rbf"))
    ]
    return cml.construction("conceptual", algorithms)


@patch("cml.domain.construction.Constructor")
def test_constructor_calling_constructor(mock_constructor):
    # arrage
    mode = "conceptual"
    ml_models = [MachineLearningModel(), MachineLearningModel()]
    settings = specific_settings_factory("construction")

    # act
    Constructor(mode, ml_models, settings)

    # assert
    assert mock_constructor.called_with(mode, ml_models, settings)


def test_constructor_proper_initialization():
    # arrange
    mode = "conceptual"
    ml_models = [MachineLearningModel(), MachineLearningModel()]
    settings = specific_settings_factory("construction")

    # act
    constructor = Constructor(mode, ml_models, settings)

    # assert
    assert constructor.mode == "conceptual"
    assert constructor.settings == settings
    assert constructor.ml_models == ml_models
    assert constructor.logger is None
    assert constructor._Constructor__process is None
    assert constructor._Constructor__input_queue is None
    assert constructor._Constructor__output_queue is None


def test_constructor_change_settings_by_users():
    # arrange
    mode = "conceptual"
    ml_models = [MachineLearningModel(), MachineLearningModel()]
    settings = specific_settings_factory("construction")
    categorical_complexity = 33
    min_category_size = 666


    # act
    constructor = Constructor(mode, ml_models, settings)
    constructor.max_categories = categorical_complexity
    constructor.min_category_size = min_category_size

    # assert
    assert constructor.max_categories == categorical_complexity
    assert constructor.min_category_size == min_category_size


def test_construction_success_with_kmeans():
    # arrange
    from sklearn.cluster import KMeans

    learnblock = PandasBlock.read_csv("./test_toyset.csv")
    algorithms = [
        ("Kme", KMeans())
    ]
    constructor = cml.construction("conceptual", algorithms)
    constructor.min_category_size = 1
    constructor.max_categories = 5

    # act
    constructed = constructor.construct(learnblock)

    # assert
    construct = iter(constructed)
    assert len(Counter(next(construct).get_column("Z")).keys()) == 2
    assert len(Counter(next(construct).get_column("Z")).keys()) == 3
    assert len(Counter(next(construct).get_column("Z")).keys()) == 4
    assert len(Counter(next(construct).get_column("Z")).keys()) == 5


def test_construction_success_with_spectral_clustering():
    # arrange
    from sklearn.cluster import SpectralClustering

    learnblock = PandasBlock.read_csv("./test_toyset.csv")
    algorithms = [
        ("Spe", SpectralClustering())
    ]
    constructor = cml.construction("conceptual", algorithms)
    constructor.min_category_size = 1
    constructor.max_categories = 5

    # act
    constructed = constructor.construct(learnblock)

    # assert
    construct = iter(constructed)
    assert len(Counter(next(construct).get_column("Z")).keys()) == 2
    assert len(Counter(next(construct).get_column("Z")).keys()) == 3
    assert len(Counter(next(construct).get_column("Z")).keys()) == 4
    assert len(Counter(next(construct).get_column("Z")).keys()) == 5


def test_construction_success_agglomerative_clustering_complete_linkage():
    # arrange
    from sklearn.cluster import AgglomerativeClustering

    learnblock = PandasBlock.read_csv("./test_toyset.csv")
    algorithms = [
        ("Agg", AgglomerativeClustering(linkage="complete"))
    ]
    constructor = cml.construction("conceptual", algorithms)
    constructor.min_category_size = 1
    constructor.max_categories = 5

    # act
    constructed = constructor.construct(learnblock)

    # assert
    construct = iter(constructed)
    assert len(Counter(next(construct).get_column("Z")).keys()) == 2
    assert len(Counter(next(construct).get_column("Z")).keys()) == 3
    assert len(Counter(next(construct).get_column("Z")).keys()) == 4
    assert len(Counter(next(construct).get_column("Z")).keys()) == 5


def test_construction_success_agglomerative_clustering_ward_linkage():
    # arrange
    from sklearn.cluster import AgglomerativeClustering

    learnblock = PandasBlock.read_csv("./test_toyset.csv")
    algorithms = [
        ("Agg", AgglomerativeClustering(linkage="ward"))
    ]
    constructor = cml.construction("conceptual", algorithms)
    constructor.min_category_size = 1
    constructor.max_categories = 5

    # act
    constructed = constructor.construct(learnblock)

    # assert
    construct = iter(constructed)
    assert len(Counter(next(construct).get_column("Z")).keys()) == 2
    assert len(Counter(next(construct).get_column("Z")).keys()) == 3
    assert len(Counter(next(construct).get_column("Z")).keys()) == 4
    assert len(Counter(next(construct).get_column("Z")).keys()) == 5


def test_construction_success_agglomerative_clustering_average_linkage():
    # arrange
    from sklearn.cluster import AgglomerativeClustering

    learnblock = PandasBlock.read_csv("./test_toyset.csv")
    algorithms = [
        ("Agg", AgglomerativeClustering(linkage="average"))
    ]
    constructor = cml.construction("conceptual", algorithms)
    constructor.min_category_size = 1
    constructor.max_categories = 5

    # act
    constructed = constructor.construct(learnblock)

    # assert
    construct = iter(constructed)
    assert len(Counter(next(construct).get_column("Z")).keys()) == 2
    assert len(Counter(next(construct).get_column("Z")).keys()) == 3
    assert len(Counter(next(construct).get_column("Z")).keys()) == 4
    assert len(Counter(next(construct).get_column("Z")).keys()) == 5


def test_construction_success_agglomerative_clustering_single_linkage():
    # arrange
    from sklearn.cluster import AgglomerativeClustering

    learnblock = PandasBlock.read_csv("./test_toyset.csv")
    algorithms = [
        ("Agg", AgglomerativeClustering(linkage="single"))
    ]
    constructor = cml.construction("conceptual", algorithms)
    constructor.min_category_size = 1
    constructor.max_categories = 5

    # act
    constructed = constructor.construct(learnblock)

    # assert
    construct = iter(constructed)
    assert len(Counter(next(construct).get_column("Z")).keys()) == 2
    assert len(Counter(next(construct).get_column("Z")).keys()) == 3
    assert len(Counter(next(construct).get_column("Z")).keys()) == 4
    assert len(Counter(next(construct).get_column("Z")).keys()) == 5


def test_construction_success_dbscan():
    # arrange
    from sklearn.cluster import DBSCAN

    learnblock = PandasBlock.read_csv("./test_toyset.csv")
    algorithms = [
        ("Dbc", DBSCAN())
    ]
    constructor = cml.construction("conceptual", algorithms)
    constructor.min_category_size = 1
    constructor.max_categories = 5

    # act
    constructed = constructor.construct(learnblock)

    # assert
    for construct in constructed:
        assert len(Counter(construct.get_column("Z")).keys()) == construct.n_cluster


def test_construction_success_birch():
    # arrange
    from sklearn.cluster import Birch

    learnblock = PandasBlock.read_csv("./test_toyset.csv")
    algorithms = [
        ("Bir", Birch())
    ]
    constructor = cml.construction("conceptual", algorithms)
    constructor.min_category_size = 1
    constructor.max_categories = 5

    # act
    constructed = constructor.construct(learnblock)

    # assert
    construct = iter(constructed)
    assert len(Counter(next(construct).get_column("Z")).keys()) == 2
    assert len(Counter(next(construct).get_column("Z")).keys()) == 3
    assert len(Counter(next(construct).get_column("Z")).keys()) == 4
    assert len(Counter(next(construct).get_column("Z")).keys()) == 5


def test_construction_optics():
    # arrange
    from sklearn.cluster import OPTICS

    learnblock = PandasBlock.read_csv("./test_toyset.csv")
    algorithms = [
        ("Opt", OPTICS())
    ]
    constructor = cml.construction("conceptual", algorithms)
    constructor.min_category_size = 1
    constructor.max_categories = 5

    # act
    constructed = constructor.construct(learnblock)

    # assert
    for construct in constructed:
        assert len(Counter(construct.get_column("Z")).keys()) == construct.n_cluster


def test_construction_correct_learnblock_label_with_subject_and_n_cluster():
    from sklearn.cluster import KMeans

    # arrange
    learnblock = PandasBlock.read_csv("./test_toyset.csv")
    algorithm = [
        ("Kme", KMeans())
    ]
    constructor = cml.construction("conceptual", algorithm)
    constructor.max_categories = 2
    constructor.min_category_size = 1

    # act
    constructed = constructor.construct(learnblock)

    # assert
    for labeled in constructed:
        print(labeled.subject)
        assert labeled.n_cluster == 2
        assert labeled.subject == "Kme02"
        assert len(Counter(labeled.get_column("Z")).keys()) == 2
        break
    else:
        assert False


def conceptual_construction_correct_logging_info():
    from sklearn.cluster import KMeans
    from cml.shared.logger import ConstructionInfo

    # arrange
    learnblock = PandasBlock.read_csv("./test_toyset.csv")
    algorithm = [
        ("Kme", KMeans())
    ]
    constructor = cml.construction("conceptual", algorithm)
    constructor.max_categories = 2

    # act
    constructed = constructor._construct_conceptual_knowledge(learnblock)

    # assert
    for log_info, labeled in constructed:
        assert log_info.state == ConstructionInfo.State.CONSTRUCTED
        assert log_info.subject == "Kme"
        assert log_info.n_cluster == 2
