import cml
import pandas as pd
from unittest.mock import MagicMock
from time import time
from cml.domain.reconstruction import Metadata
from cml.ports.source_adapter import PandasBlock
from cml.domain.reconstruction import (
    split,
    calc_reliability,
    valid_reconstructed,
    PragmaticMachineLearningModel,
    determine_winner)


DF = pd.DataFrame(
    {
       "0.0.1": [pow(i, 2) for i in range(10, 100)],
       "0.0.2": [pow(i, 0.5) for i in range(10, 100)],
       "0.0.3": [pow(i, 3) for i in range(10, 100)],
       "0.0.4": [pow(i, 0.3) for i in range(10, 100)],
       "0.0.5": [pow(i, 0.1) for i in range(10, 100)],
       "0.0.6": [pow(i, 4) for i in range(10, 100)],
       "T": [int(time()) for i in range(10, 100)],
       "Sigma": ["" for i in range(10, 100)],
       "Z": [0 if i < 50 else 1 for i in range(10, 100)]
    }
)


def get_labeled_block():
    labeled = PandasBlock(DF, relationship=("Z", "Sigma"), origin=("source", ))
    labeled.subject = "Kme02"
    labeled.n_cluster = 2
    return labeled


def test_learnblock_splitting_for_training():
    # act
    labeled = get_labeled_block()
    training_block, evaluation_block = split(labeled, 0.2)
    training_data_size = int(labeled.n_rows() * 0.8)
    evaluation_data_size = int(labeled.n_rows()-training_data_size)

    # assert
    assert training_block.n_rows() == training_data_size
    assert evaluation_block.n_rows() == evaluation_data_size


def test_calc_reliability():
    from sklearn.svm import SVC
    # arrange
    labeled = get_labeled_block()
    mode = "conceptual"
    algos = [("Svc", SVC())]
    reconstructor = cml.reconstruction(mode, algos)
    reconstructor.next_model_counter = lambda x: x+1
    pragmatic = reconstructor.reconstruct(tier=1, learnblock=labeled)
    _, evaluation_block = split(labeled, 0.2)

    # act
    alpha_coefficient = calc_reliability(pragmatic.model, evaluation_block,
                                         "ratio")

    # assert
    assert alpha_coefficient >= 0
    assert isinstance(alpha_coefficient, float)


def test_valid_reconstructed_success_false():
    from sklearn.svm import SVC

    # arrange
    labeled = get_labeled_block()
    mode = "conceptual"
    algos = [("Svc", SVC())]
    reconstructor = cml.reconstruction(mode, algos)
    reconstructor.next_model_counter = lambda x: x+1
    pragmatic = reconstructor.reconstruct(tier=1, learnblock=labeled)

    _, evaluation_block = split(labeled, 0.2)
    pragmatic.model.accuracy = 0.5

    # act
    response = valid_reconstructed(pragmatic.model, "conceptual", 0.6, None, None)

    # assert
    assert response == False


def test_valid_reconstructed_success_true():
    from sklearn.svm import SVC
    # arrange
    labeled = get_labeled_block()
    mode = "conceptual"
    algos = [("Svc", SVC())]
    reconstructor = cml.reconstruction(mode, algos)
    reconstructor.next_model_counter = lambda x: x+1
    pragmatic = reconstructor.reconstruct(tier=1, learnblock=labeled)

    _, evaluation_block = split(labeled, 0.2)
    pragmatic.model.accuracy = 0.7

    # act
    response = valid_reconstructed(pragmatic.model, "conceptual", 0.6, None, None)

    # assert
    assert response == True


def test_conceptual_reconstructor_creation():
    from sklearn.svm import SVC

    # arrange
    labeled = get_labeled_block()
    mode = "conceptual"
    algos = [
        ("Svc", SVC())
    ]
    min_t = min(DF["T"])
    max_t = max(DF["T"])

    # act
    reconstructor = cml.reconstruction(mode, algos)
    reconstructor.next_model_counter = lambda x: x+1
    pragmatic = reconstructor.reconstruct(tier=1, learnblock=labeled)

    # assert
    assert pragmatic.uid == "C.1.2"
    assert pragmatic.min_timestamp == min_t
    assert pragmatic.max_timestamp == max_t
    assert pragmatic.subject == ("Svc", )
    assert pragmatic.purpose == ("C.1.Kme02", )


def test_conceptual_reconstruction_with_many_models():
    from sklearn.svm import SVC
    from sklearn.linear_model import LinearRegression, BayesianRidge, ElasticNet

    # arrange
    labeled = get_labeled_block()
    mode = "conceptual"
    algos = [
        ("Svc", SVC()),
        ("LiR", LinearRegression()),
        ("BaR", BayesianRidge()),
        ("Ela", ElasticNet())
    ]

    # act
    reconstructor = cml.reconstruction(mode, algos)
    reconstructor.next_model_counter = lambda x: x+1
    pragmatic = reconstructor.reconstruct(tier=1, learnblock=labeled)

    # assert
    assert pragmatic.uid == "C.1.2"
    assert pragmatic.subject in [("Svc", ), ("Lir", ), ("BaR", ), ("Ela", )]


def test_conceptual_reconstruction_with_meta():
    from sklearn.svm import SVC
    from sklearn.linear_model import LinearRegression

    # arrange
    labeled = PandasBlock(DF, relationship=("Z", "Sigma"), origin=("C.1.2", ))
    labeled.subject = "Kme02"
    labeled.n_cluster = 2
    mode = "conceptual"
    algos = [("Svc", SVC()), ("LiR", LinearRegression())]
    min_t = labeled.min_timestamp()
    max_t = labeled.max_timestamp()
    meta = Metadata(
        "C", 2, 4, None, None, None, min_t, max_t,
        ("Svc", ), ("C.1.Kme02", "C.2.Kmea02"))

    # act
    reconstructor = cml.reconstruction(mode, algos)
    reconstructor.next_model_counter = lambda x: x+1
    pragmatic = reconstructor.reconstruct(
        tier=2, learnblock=labeled, which_models=["Svc"], meta=meta)

    # assert
    assert pragmatic.uid == "C.2.4"
    assert pragmatic.tier == 2
    assert pragmatic.min_timestamp == min_t
    assert pragmatic.max_timestamp == max_t
    assert pragmatic.subject == meta.sigma
    assert pragmatic.purpose == meta.zeta
    assert pragmatic.counter == 4
    assert pragmatic.learnblock == labeled


def test_determine_winner_from_one():
    # arrange
    p_mock_one = MagicMock(spec=PragmaticMachineLearningModel)
    p_mock_one.origin = ()
    p_mock_two = MagicMock(spec=PragmaticMachineLearningModel)
    p_mock_two.origin = ()
    p_mock_three = MagicMock(spec=PragmaticMachineLearningModel)
    p_mock_three.origin = ()

    sorted_reliabilities = {
        0.921: [p_mock_one],
        0.431: [p_mock_two],
        0.999: [p_mock_three]
    }

    # act
    winner = determine_winner(sorted_reliabilities)

    # assert
    assert winner == sorted_reliabilities[0.999].pop()


def test_determine_winner_from_many():
    # arrange
    p_mock_one = MagicMock(spec=PragmaticMachineLearningModel)
    p_mock_one.origin = ()
    p_mock_two = MagicMock(spec=PragmaticMachineLearningModel)
    p_mock_two.origin = ()
    p_mock_three = MagicMock(spec=PragmaticMachineLearningModel)
    p_mock_three.origin = ()
    p_mock_three.pre_image_features = [1, 2, 3, 4]
    p_mock_four = MagicMock(spec=PragmaticMachineLearningModel)
    p_mock_four.origin = ()
    p_mock_four.pre_image_features = [1, 2, 3, 4, 5, 6]

    sorted_reliabilities = {
        0.921: [p_mock_one],
        0.431: [p_mock_two],
        0.999: [p_mock_three, p_mock_four]
    }

    # act
    winner = determine_winner(sorted_reliabilities)

    # assert
    assert winner == sorted_reliabilities[0.999][0]
